note
	description: "Summary description for {TEST_LOOP}."
	author: "Jackie Wang"
	date: "$Date$"
	revision: "$Revision$"

class
	TEST_LOOP

inherit
	ES_TEST

create
	make

feature -- Loops

	find_max (a: ARRAY [INTEGER]): INTEGER
		local
			i: INTEGER
		do
			from
				i := a.lower
				Result := a [i]
			invariant
				correct_loop_invariant:
					across a.lower |..| (i - 1) as j
						all Result >= a [j.item] end
			until
				i > a.upper
			loop
				if a [i] > Result then
					Result := a [i]
				end
				i := i + 1
			variant
				correct_variant:
					a.upper - i + 1
			end
		ensure
			across a.lower |..| a.upper as j
				all Result >= a [j.item]
			end
		end

	find_max_wrong_invariant (a: ARRAY [INTEGER]): INTEGER
		local
			i: INTEGER
		do
			from
				i := a.lower
				Result := a [i]
			invariant
				-- try this failing loop invariant, why?
				wrong_loop_invariant:
					across a.lower |..| i as j
						all Result >= a [j.item] end
			until
				i > a.upper
			loop
				if a [i] > Result then
					Result := a [i]
				end
				i := i + 1
			variant
				correct_loop_variant:
					a.upper - i + 1
			end
		ensure
			across a.lower |..| a.upper as j
				all Result >= a [j.item]
			end
		end

	find_max_wrong_variant (a: ARRAY [INTEGER]): INTEGER
		local
			i: INTEGER
		do
			from
				i := a.lower
				Result := a [i]
			invariant
				correct_loop_invariant:
					across a.lower |..| (i - 1) as j
						all Result >= a [j.item] end
			until
				i > a.upper
			loop
				if a [i] > Result then
					Result := a [i]
				end
				i := i + 1
			variant
				-- try this failing variant, why?
				 wrong_loop_variant:
				 	a.upper - i
			end
		ensure
			across a.lower |..| a.upper as j
				all Result >= a [j.item]
			end
		end

feature -- Constructor

	make
		do
			add_boolean_case (agent test_correct_loop_invariant_variant)
			add_violation_case_with_tag ("wrong_loop_invariant", agent test_wrong_loop_invariant)
			add_violation_case_with_tag ("wrong_loop_variant", agent test_wrong_loop_variant)

			show_browser
			run_espec
		end

feature -- Tests

	test_correct_loop_invariant_variant: BOOLEAN
		local
			max: INTEGER
		do
			comment ("test_correct_loop_invariant_variant: loop with correct invariant and variant")
			max := find_max (<<20, 10, 40, 30>>)
			Result := max = 40
			check Result end
		end

	test_wrong_loop_invariant
		local
			max: INTEGER
		do
			comment ("test_wrong_loop_invariant: loop with wrong variant")
			max := find_max_wrong_invariant (<<20, 10, 40, 30>>)
		end

	test_wrong_loop_variant
		local
			max: INTEGER
		do
			comment ("test_wrong_loop_variant: loop with wrong variant")
			max := find_max_wrong_variant (<<20, 10, 40, 30>>)
		end
end
